"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const os_1 = require("os");
const worker_threads_1 = require("worker_threads");
const assertExists_1 = __importDefault(require("./assertExists"));
const Dict_1 = __importDefault(require("./Dict"));
const UnreachableError_1 = __importDefault(require("./UnreachableError"));
class TaskIO {
    taskHandlers;
    sentRequests = new Dict_1.default();
    nextSendRequestId = 0;
    constructor(taskHandlers) {
        this.taskHandlers = taskHandlers;
    }
    handleMessage(ctx, port, msg) {
        if (msg.$type === "response") {
            const { id, output } = msg;
            const { resolve } = this.sentRequests.remove(id);
            resolve(output);
        }
        else if (msg.$type === "request") {
            const { id, type, input } = msg;
            const handler = this.taskHandlers[type];
            // TODO Handle errors.
            handler(input, ctx).then((output) => {
                port.postMessage({
                    $type: "response",
                    id,
                    output,
                });
            });
        }
        else {
            throw new UnreachableError_1.default();
        }
    }
    sendRequest(to, type, input) {
        return new Promise((resolve) => {
            const id = this.nextSendRequestId++;
            this.sentRequests.set(id, { resolve });
            to.postMessage({ $type: "request", id, type, input });
        });
    }
}
class WorkerPool {
    scriptPath;
    n;
    stateInit;
    entrypoint;
    taskHandlers = {};
    constructor(scriptPath, n = (0, os_1.cpus)().length) {
        this.scriptPath = scriptPath;
        this.n = n;
    }
    leader(fn) {
        if (worker_threads_1.isMainThread) {
            this.entrypoint = fn;
        }
        return this;
    }
    worker(fn) {
        if (!worker_threads_1.isMainThread) {
            this.entrypoint = fn;
        }
        return this;
    }
    leaderState(fn) {
        if (worker_threads_1.isMainThread) {
            this.stateInit = fn;
        }
        return this;
    }
    workerState(fn) {
        if (!worker_threads_1.isMainThread) {
            this.stateInit = fn;
        }
        return this;
    }
    leaderTask(name, handler) {
        if (worker_threads_1.isMainThread) {
            this.taskHandlers[name] = handler;
        }
        return this;
    }
    workerTask(name, handler) {
        if (!worker_threads_1.isMainThread) {
            this.taskHandlers[name] = handler;
        }
        return this;
    }
    async go() {
        const state = await this.stateInit?.();
        const io = new TaskIO(this.taskHandlers);
        if (worker_threads_1.isMainThread) {
            // We're inside the leader.
            const workers = Array.from({ length: this.n }, () => new worker_threads_1.Worker(this.scriptPath));
            // TODO Handle `w.on("error")`.
            for (const w of workers) {
                w.addListener("message", (msg) => io.handleMessage(ctx, w, msg));
            }
            let nextWorkerId = 0;
            const ctx = {
                state,
                execute: (type, input) => {
                    const worker = workers[(nextWorkerId = (nextWorkerId + 1) % workers.length)];
                    return io.sendRequest(worker, type, input);
                },
            };
            this.entrypoint?.(ctx);
        }
        else {
            // We're inside a worker.
            const leader = (0, assertExists_1.default)(worker_threads_1.parentPort);
            leader.addListener("message", (msg) => io.handleMessage(ctx, leader, msg));
            const ctx = {
                state,
                execute: (type, input) => io.sendRequest(leader, type, input),
            };
            // Does not have to exist.
            this.entrypoint?.(ctx);
        }
    }
}
exports.default = WorkerPool;
